// V810 0.5 Recognizer Grammar
//
// Run 'java V810Asm <V810 file>'
//
// This grammar is in the PUBLIC DOMAIN
// (C) 1998 David Tucker & the VB-98 team...
//  dbt@dana.ucc.nau.edu
//  http://dana.ucc.nau.edu/~dbt/VBMain.html
//

// Import the necessary classes
{
import java.io.*;
}


//-----------------------------------------------------------------------------
// Define a Parser, calling it V810Asm
//-----------------------------------------------------------------------------
class V810Asm extends Parser;
options {
	k = 2;                           // two token lookahead
	tokenVocabulary=V810;            // Call its vocabulary "V810"
	codeGenMakeSwitchThreshold = 2;  // Some optimizations
	codeGenBitsetTestThreshold = 3;
	defaultErrorHandler = false;     // Don't generate parser error handlers
}


// Define some methods and variables to use in the generated parser.
{
	// Define a generica array that can hold our Instruction and modifiers, along
	// with some extra details...
	// format: 
	//   0) 'opcode type' (10,20,30,40,50,60,61,70,80,etc...)
	//   1) full Instruction
	//   2) opcode (00000..11111)
	//   3)var1
	//   4)var2
	//   5)var3
	//   6)etc...
	private int[] op_format = {0,0,0,0,0,0,0,0}; // array of 8, for padding...

	// Buffer to hold DB. valuse...
	private int[] defBuff = new int[256];

	// PC, used to help .org defines
	long tmp_PC = 0;

	//The OBJ Output
	private static RandomAccessFile dout = null;

	//Define a Inst type
	// Define a main
	public static void main(String[] args) { //static
                
		// Use a try/catch block for parser exceptions
		try {
			// if we have at least one command-line argument
			if (args.length > 0 ) {
				System.err.println("Parsing...");

				// for each directory/file specified on the command line
				for(int i=0; i< args.length;i++)
					doFile(new File(args[i])); // parse it
			} else {
				System.err.println("Usage: java V810Asm "+"<file>");
			}
		} catch(Exception e) {
			System.err.println("exception: "+e);
			e.printStackTrace(System.err);   // so we can get stack trace
		}
	}

	// This method decides what action to take based on the type of
	//   file we are looking at
	public static void doFile(File f) throws Exception {
		// If this is a directory, Return, were not smart enoug to handle them =0)
		if (f.isDirectory()) return;

		// otherwise, if this is a V810 file, parse it!
		else if ((f.getName().length()>4) && f.getName().substring(f.getName().length()-4).equals(".asm")) {
			System.err.println("   "+f.getAbsolutePath());
			//Open our output file....
			File obj = new File(f.getName().substring(0,f.getName().length()-4)+".obj");
			dout = new RandomAccessFile(obj, "rw");

			parseFile(new FileInputStream(f));
                        
			dout.close();
		}
	}

	// Here's where we do the real work...
	public static void parseFile(InputStream s) throws Exception {
		try {
			// Create a scanner that reads from the input stream passed to us
			V810Lexer lexer = new V810Lexer(s);

			// Create a parser that reads from the scanner
			V810Asm parser = new V810Asm(lexer);

			// start parsing at the compilationUnit rule
			parser.compilationUnit();
		} catch (Exception e) {
			System.err.println("parser exception: "+e);
			e.printStackTrace();   // so we can get stack trace             
		}
	}
}

//--------------------------------------------------------------
// Compilation Unit: In V810, this is a single file.  This is the start
//   rule for this parser
//--------------------------------------------------------------
compilationUnit
	:	// A compilation unit contains preprocesor code and asm code
		tmp_PC=org { tmp_PC = (tmp_PC & 0xFFFFFFFC);} // Init the PC
		(       preProc
		|       code
		)*
		EOF
	;

//--------------------------------------------------------------
// PreProcessor code Undefined for now...
//--------------------------------------------------------------
preProc
{long tOrg = 0;}
	: defNum   // Up to 256 dnum's max...
	| defStr   // same hear...
	//|alocateNum, Needed???
	//defLiteral, ie dlb. dlw. dlh. (No String)
	| tOrg=org	{ // Inlined Java Code, fills in the OBJ with zeros, to the new .org (Bad form)
		if (tOrg < tmp_PC+dout.getFilePointer()) {
			System.out.println("Error: org("+tOrg+") < CurPC("+(tmp_PC+dout.getFilePointer())+")");
		} else {
			while(tOrg > tmp_PC+dout.getFilePointer()) {
				dout.writeByte(0xFF);
			}
		}
	  } //End Inline
	;

org returns [int tmp]
{tmp = -1;}
	: "org" DOT tmp=num {if(tmp < 0) {
		System.out.println("Error, org. is invalid: " + tmp); tmp = 0;} 
	  }
	;


defNum
{int count = 0;}
	: "db" DOT count=dnum {
		for(int i=0; i<count; i++) {      // Define Byte
			dout.writeByte((byte)(defBuff[i]&0xFF));
		}
	  }
	| "dh" DOT count=dnum {
		for(int i=0; i<count; i++) {      // Defien HWord
			dout.writeByte((byte)(defBuff[i]&0xFF));
			dout.writeByte((byte)((defBuff[i]>>8)&0xFF));
		}
	  }
	| "dw" DOT count=dnum {
		for(int i=0; i<count; i++) {      // Define Word
			dout.writeByte((byte)(defBuff[i]&0xFF));
			dout.writeByte((byte)((defBuff[i]>>8)&0xFF));
			dout.writeByte((byte)((defBuff[i]>>16)&0xFF));
			dout.writeByte((byte)((defBuff[i]>>24)&0xFF));
		}
	  }
	;

defStr	// I should be using writeUTF, but there is a bug in it, it adds garbage chars...
	: "ds" DOT s:STR {dout.writeBytes(s.getText());} // Define String  //UTF
	;

dnum returns [int count]
{
	int ret = 0;
	count = 0;
}
	: ret=num {defBuff[count++]=ret;}
		(COMMA ret=num {defBuff[count++]=ret;} )*
	;

//-------------------------------------------------------------
// code is an opcode with 0 or more variables
//-------------------------------------------------------------
code
options {defaultErrorHandler = true;}
{int n = 0;}
	:	n=opp {op_format[2] = n;} //{System.out.println(" opp: " + n);}
		vars
		{
			op_format[1]=0; // Clear the opcode...
			//Case off of Op_Format[0]
			switch(op_format[0]) {
				case 10:        //Mode1, 2, 9
				case 20:
				case 90:
					op_format[1] = (((op_format[2]&0x3F) << 10) + ((op_format[4]&0x1F) << 5) + (op_format[3]&0x1F));
					dout.writeByte((byte)(op_format[1] &0xFF));
					dout.writeByte((byte)((op_format[1] >> 8) &0xFF));
					break;
				case 30:        //Mode3
					op_format[1] = (((op_format[2]&0x7F) << 9) + (op_format[3]&0x1FF));
					dout.writeByte((byte)(op_format[1] &0xFF));
					dout.writeByte((byte)((op_format[1] >> 8) &0xFF));
				break;
					case 40:        //Mode4
					op_format[1] = (((op_format[2]&0x3F) << 26) + (op_format[3]&0x3FFFFFF));
					dout.writeByte((byte)((op_format[1] >> 16) &0xFF));
					dout.writeByte((byte)((op_format[1] >> 24) &0xFF));

					dout.writeByte((byte)(op_format[1]         &0xFF));
					dout.writeByte((byte)((op_format[1] >>  8) &0xFF));
					break;
				case 50:        //Mode5, 6a
				case 60:
					op_format[1] = (((op_format[2]&0x3F) << 26) + ((op_format[5]&0x1F) << 21) + ((op_format[4]&0x1F)<< 16) + (op_format[3]&0xFFFF));
					dout.writeByte((byte)((op_format[1] >> 16) &0xFF));
					dout.writeByte((byte)((op_format[1] >> 24) &0xFF));

					dout.writeByte((byte)(op_format[1]         &0xFF));
					dout.writeByte((byte)((op_format[1] >>  8) &0xFF));
					break;
				case 61:        //Mode6b
					op_format[1] = (((op_format[2]&0x3F) << 26) + ((op_format[3]&0x1F) << 21) + ((op_format[5]&0x1F)<< 16) + (op_format[4]&0xFFFF));
					dout.writeByte((byte)((op_format[1] >> 16) &0xFF));
					dout.writeByte((byte)((op_format[1] >> 24) &0xFF));

					dout.writeByte((byte)(op_format[1]         &0xFF));
					dout.writeByte((byte)((op_format[1] >>  8) &0xFF));
					break;
				default:        //Ops, howd we get hear....
					break;
			}
			System.out.print("."); //Limit this more...
		}
	;
        

//-------------------------------------------------------------
// Vars = 0 or more variables
//-------------------------------------------------------------
vars
{
	int n = 0;
	int count = 0;
	op_format[3]=0; //Zero the vars...
	op_format[4]=0;
	op_format[5]=0;
}
	: /* No var break recursion*/
	| n=var {op_format[3+count] = n; count ++; } // {count +=1; System.out.println(" var" + count + ": " + n);} //1'st num/reg
	  vars2[count]
	| LBRACK n=var RBRACK {op_format[3+count] = n; count ++; } // {count +=1; System.out.println(" var" + count + ": " + n);} // Breacketed reg
	  vars2[count]
	;
vars2[int count]
{int n = 0;}
	: /* No var break recursion*/
	| COMMA n=var {op_format[3+count] = n; count ++; } // {count +=1; System.out.println(" var" + count + ": " + n);} // remaning num/reg
	  vars2[count]
	| LBRACK n=var RBRACK {op_format[3+count] = n; count ++; } // {count +=1; System.out.println(" var" + count + ": " + n);} // Breacketed reg
	  vars2[count]
	;
// variable is Ither a num or a register, or a Inderect refrence
var returns [int n]
	options {defaultErrorHandler = true;}
{n = 0;}
	:       n=reg
	//|     n=ref
	|       n=num
	;

//  Define a register... '$3'
reg returns [int n]
{n=0;}
	:       DOLAR n=num 
	;

// Define our Opcodes
protected
opp returns [int g]
{
	g = 0;
	int tmp = 0;
}

// MODE1
	:(
			"sub"		{g = 0x02;}
		|	"jmp"		{g = 0x06;}
		|	"mul"		{g = 0x08;}
		|	"div"		{g = 0x09;}
		|	"mulu"	{g = 0x0A;}
		|	"divu"	{g = 0x0B;}
		|	"or"		{g = 0x0C;}
		|	"and"		{g = 0x0D;}
		|	"xor"		{g = 0x0E;}
		|	"not"		{g = 0x0F;}
	) {op_format[0] = 10;}

// MODE1 & 2 conflicts...
	|(
			("mov" DOLAR)=> "mov"	{g = 0x00; op_format[0] = 10;}
		|	("add" DOLAR)=> "add"	{g = 0x01; op_format[0] = 10;} // COMMA tmp=reg 
		|	("cmp" DOLAR)=> "cmp"	{g = 0x03; op_format[0] = 10;}
		|	("shl" DOLAR)=> "shl"	{g = 0x04; op_format[0] = 10;}
		|	("shr" DOLAR)=> "shr"	{g = 0x05; op_format[0] = 10;}
		|	("sar" DOLAR)=> "sar"	{g = 0x07; op_format[0] = 10;}

		|	"mov"		{g = 0x10; op_format[0] = 20;} // Imediate 
		|	"add"		{g = 0x11; op_format[0] = 20;}
		|	"cmp"		{g = 0x13; op_format[0] = 20;}
		|	"shl"		{g = 0x14; op_format[0] = 20;}
		|	"shr"		{g = 0x15; op_format[0] = 20;}
		|	"sar"		{g = 0x17; op_format[0] = 20;}
	)               

//MODE2
	|(
			"setf"	{g = 0x12;}
		|	"ei"		{g = 0x16;}
		|	"trap"	{g = 0x18;}
		|	"ldsr"	{g = 0x1C;}
		|	"stsr"	{g = 0x1D;}
		|	"di"		{g = 0x1E;}
	) {op_format[0] = 20;}

//MODE3
	|(
			"bv"		{g = 0x40;}
		|	"bl"		{g = 0x41;} //BC  0x41 
		|	"be"		{g = 0x42;} //BZ  0x42 
		|	"bnh"		{g = 0x43;}
		|	"bn"		{g = 0x44;}
		|	"br"		{g = 0x45;}
		|	"blt"		{g = 0x46;}
		|	"ble"		{g = 0x47;}
		|	"bnv"		{g = 0x48;}
		|	"bnl"		{g = 0x49;} //BNC 0x49 
		|	"bne"		{g = 0x4A;} //BNZ 0x4A 
		|	"bh"		{g = 0x4B;}
		|	"bp"		{g = 0x4C;}
		|	"nop"		{g = 0x4D;}
		|	"bge"		{g = 0x4E;}
		|	"bgt"		{g = 0x4F;}
	) {op_format[0] = 30;}

//MODE4
	|(
			"jr"		{g = 0x2A;}
		|	"jal"		{g = 0x2B;}
	) {op_format[0] = 40;}

//MODE5
	|(
			"movea"	{g = 0x28;}
		|	"addi"	{g = 0x29;}
		|	"ori"		{g = 0x2C;}
		|	"andi"	{g = 0x2D;}
		|	"xori"	{g = 0x2E;}
		|	"movhi"	{g = 0x2F;}
	) {op_format[0] = 50;}

//MODE6A
	|(
			"ld" DOT (
				 "b"	{g = 0x30;}
				|"h"	{g = 0x31;}
				|"w"	{g = 0x33;}
			)
		|	"in" DOT (
				 "b"	{g = 0x38;}
				|"h"	{g = 0x39;}
				|"w"	{g = 0x3B;}
			)
		|	"caxi"	{g = 0x3A;}
	) {op_format[0] = 60;}

//MODE6B
	|(
			"st" DOT (
				 "b"	{g = 0x34;}
				|"h"	{g = 0x35;}
				|"w"	{g = 0x37;}
			)
		|	"out" DOT (
				 "b"	{g = 0x3C;}
				|"h"	{g = 0x3D;}
				|"w"	{g = 0x3F;}
			)
	) {op_format[0] = 61;}

//MODE9
	|(
			"reti"  {g = 0x19;} //BRKRETI 
		|	"halt"  {g = 0x1A;} //STBY 
	) {op_format[0] = 90;}
	;

//Junk
	/*:	"???"  {AM_UDEF,x1B } // Unknown*/
	/*|	"BSTR" {AM_BSTR,x1F } // Special Bit String Instructions*/
	/*|	"???"  {AM_UDEF,x20 } // Unknown*/ // This is a fudg on our part
	/*|	"???"  {AM_UDEF,x21 } // Unknown*/ // We have 6 and 7 bit instructions
	/*|	"???"  {AM_UDEF,x22 } // Unknown*/ // this is filld in by the Conditional Branch Instructions
	/*|	"???"  {AM_UDEF,x23 } // Unknown*/
	/*|	"???"  {AM_UDEF,x24 } // Unknown*/
	/*|	"???"  {AM_UDEF,x25 } // Unknown*/
	/*|	"???"  {AM_UDEF,x26 } // Unknown*/
	/*|	"???"  {AM_UDEF,x27 } // Unknown*/
	/*|	"muli" {AM_UDEF,x32 } // V830*/
	/*|	"maci" {AM_UDEF,x36 } // V830*/
	/*|	"FPP"  { AM_FPP,x3E } // Floating Point Instruction, Special Case*/


num     returns [int n]
{n=0;}
	: i:INT {n=(Integer.valueOf(i.getText())).intValue();}
	;


//----------------------------------------------------------------------------
// The V810 scanner
//----------------------------------------------------------------------------
class V810Lexer extends Lexer;

options {
	tokenVocabulary=V810;  // call the vocabulary "V810"
	testLiterals=false;    // don't automatically test for literals
	k=4;                   // four characters of lookahead
}



// OPERATORS
LPAREN	:	'('	;
RPAREN	:	')'	;
LBRACK	:	'['	;
RBRACK	:	']'	;
LCURLY	:	'{'	;
RCURLY	:	'}'	;
COLON		:	':'	;
COMMA		:	','	;
DOT		:	'.'	;
SEMI		:	';'	;
DOLAR		:	'$'	;
QUOTE		:	'"'	;

// Whitespace -- ignored
WS
	:	(	' '
		|	'\t'
		|	'\f'
		// handle newlines
		|	(	"\r\n"  // Evil DOS
			|	'\r'    // Macintosh
			|	'\n'    // Unix (the right way)
			)
			{ newline(); }
		)
		{ _ttype = Token.SKIP; }
	;


// Single-line comments
SL_COMMENT
	:	"//" (~'\n')* '\n'
		{ _ttype = Token.SKIP; newline(); }
	;


// multiple-line comments
ML_COMMENT
	:	"/*"
		(	{ LA(2)!='/' }? '*'
		|	'\n' { newline(); }
		|	~('*'|'\n')
		)*
		"*/"
		{ _ttype = Token.SKIP; }
	;

// character literals
CHAR_LITERAL
	:	'\'' ( ESC | ~'\'' ) '\''
	;

// string literals
//STRING_LITERAL
//	:	'"' (ESC|~('"'|'\\'))* '"'
//	;

// escape sequence -- note that this is protected; it can only be called
//   from another lexer rule -- it will not ever directly return a token to
//   the parser
// There are various ambiguities hushed in this rule.  The optional
// '0'...'9' digit matches should be matched here rather than letting
// them go back to STRING_LITERAL to be matched.  ANTLR does the
// right thing by matching immediately; hence, it's ok to shut off
// the FOLLOW ambig warnings.
protected
ESC
	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'"'
		|	'\''
		|	'\\'
		|	('u')+ HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT 
		|	('0'..'3')
		(
			options {
				warnWhenFollowAmbig = false;
			}
			:	('0'..'9')
			(       
				options {
					warnWhenFollowAmbig = false;
				}
				:	'0'..'9'
			)?
		)?
		|	('4'..'7')
		(
			options {
				warnWhenFollowAmbig = false;
			}
			:	('0'..'9')
		)?
		)
	;


// a dummy rule to force vocabulary to be all characters (except special
//   ones that ANTLR uses internally (0 to 2)
protected
VOCAB
	:	'\3'..'\377'
	;


// an identifier.  Note that testLiterals is set to true!  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
IDENT
	options {testLiterals=true;}
	:	('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'_'|'0'..'9')*
	;

STR
	//options {testLiterals=true;}
	:	QUOTE! ('#'..'~'|'!'|' ')* QUOTE! //|'\t'|'\f'|'\r'!|'\n'!
	;

// hexadecimal digit (again, note it's protected!)
protected
OCT_DIGIT
	: ('0'..'7')
	;

protected
DEC_DIGIT
	: (OCT_DIGIT|'8'..'9')
	;

protected
HEX_DIGIT
	: (DEC_DIGIT|'A'..'F'|'a'..'f')
	;

INT
{int base = 0;}
	: base=INT_INT  {setText( (Integer.valueOf($getText,base)).toString() );}
	;

protected
INT_INT returns [int base] //cluge!
{base = 0;}
	: ('-' '0' ('x'|'X') HEX_DIGIT )=> '-' '0'! ('x'!|'X'!) (HEX_DIGIT)+	{base = 16;}
	| (    '0' ('x'|'X') HEX_DIGIT )=> '0'! ('x'!|'X'!) (HEX_DIGIT)+		{base = 16;}
	| ('-' '0' OCT_DIGIT )=> '-' '0'! (OCT_DIGIT)+					{base =  8;}
	| (    '0' OCT_DIGIT )=> '0'! (OCT_DIGIT)+					{base =  8;}
	| '-' ('1'..'9') (DEC_DIGIT)*								{base = 10;}
	|     ('1'..'9') (DEC_DIGIT)*								{base = 10;}
	| '0'												{base = 10;}
	;
